@@ -12,14 +12,14 @@ from typing import TYPE_CHECKING, Optional, Union, Callable, Sequence
 from electrum.storage import WalletStorage, StorageReadWriteError
 from electrum.wallet_db import WalletDB
 from electrum.wallet import Wallet, InternalAddressCorruption, Abstract_Wallet
-from electrum.wallet import update_password_for_directory
 
 from electrum.plugin import run_hook
 from electrum import util
 from electrum.util import (profiler, InvalidPassword, send_exception_to_crash_reporter,
                            format_satoshis, format_satoshis_plain, format_fee_satoshis,
-                           maybe_extract_bolt11_invoice, parse_max_spend)
-from electrum.invoices import PR_PAID, PR_FAILED
+                           parse_max_spend)
+from electrum.util import EventListener, event_listener
+from electrum.invoices import PR_PAID, PR_FAILED, Invoice
 from electrum import blockchain
 from electrum.network import Network, TxBroadcastError, BestEffortRequestFailed
 from electrum.interface import PREFERRED_NETWORK_PROTOCOL, ServerAddr
@@ -99,7 +99,7 @@ if TYPE_CHECKING:
     from electrum.paymentrequest import PaymentRequest
 
 
-class ElectrumWindow(App, Logger):
+class ElectrumWindow(App, Logger, EventListener):
 
     electrum_config = ObjectProperty(None)
     language = StringProperty('en')
@@ -234,10 +234,6 @@ class ElectrumWindow(App, Logger):
     def set_URI(self, uri):
         self.send_screen.set_URI(uri)
 
-    @switch_to_send_screen
-    def set_ln_invoice(self, invoice):
-        self.send_screen.set_ln_invoice(invoice)
-
     def on_new_intent(self, intent):
         data = str(intent.getDataString())
         scheme = str(intent.getScheme()).lower()
@@ -252,22 +248,28 @@ class ElectrumWindow(App, Logger):
         if self.history_screen:
             self.history_screen.update()
 
-    def on_quotes(self, d):
+    @event_listener
+    def on_event_on_quotes(self):
         self.logger.info("on_quotes")
         self._trigger_update_status()
         self._trigger_update_history()
 
-    def on_history(self, d):
+    @event_listener
+    def on_event_on_history(self):
         self.logger.info("on_history")
         if self.wallet:
             self.wallet.clear_coin_price_cache()
         self._trigger_update_history()
 
-    def on_fee_histogram(self, *args):
+    @event_listener
+    def on_event_fee_histogram(self, *args):
         self._trigger_update_history()
 
-    def on_request_status(self, event, wallet, key, status):
-        req = self.wallet.receive_requests.get(key)
+    @event_listener
+    def on_event_request_status(self, wallet, key, status):
+        if wallet != self.wallet:
+            return
+        req = self.wallet.get_request(key)
         if req is None:
             return
         if self.receive_screen:
@@ -281,7 +283,10 @@ class ElectrumWindow(App, Logger):
             self.show_info(_('Payment Received') + '\n' + key)
             self._trigger_update_history()
 
-    def on_invoice_status(self, event, wallet, key):
+    @event_listener
+    def on_event_invoice_status(self, wallet, key):
+        if wallet != self.wallet:
+            return
         req = self.wallet.get_invoice(key)
         if req is None:
             return
@@ -295,12 +300,18 @@ class ElectrumWindow(App, Logger):
         if self.invoice_popup and self.invoice_popup.key == key:
             self.invoice_popup.update_status()
 
-    def on_payment_succeeded(self, event, wallet, key):
+    @event_listener
+    def on_event_payment_succeeded(self, wallet, key):
+        if wallet != self.wallet:
+            return
         description = self.wallet.get_label(key)
         self.show_info(_('Payment succeeded') + '\n\n' + description)
         self._trigger_update_history()
 
-    def on_payment_failed(self, event, wallet, key, reason):
+    @event_listener
+    def on_event_payment_failed(self, wallet, key, reason):
+        if wallet != self.wallet:
+            return
         self.show_info(_('Payment failed') + '\n\n' + reason)
 
     def _get_bu(self):
@@ -394,10 +405,11 @@ class ElectrumWindow(App, Logger):
         self.is_exit = False
         self.wallet = None  # type: Optional[Abstract_Wallet]
         self.pause_time = 0
-        self.asyncio_loop = asyncio.get_event_loop()
+        self.asyncio_loop = util.get_asyncio_loop()
         self.password = None
         self._use_single_password = False
         self.resume_dialog = None
+        self.gui_thread = threading.current_thread()
 
         App.__init__(self)#, **kwargs)
         Logger.__init__(self)
@@ -443,12 +455,14 @@ class ElectrumWindow(App, Logger):
         self._init_finished = True
 
     def on_pr(self, pr: 'PaymentRequest'):
+        Clock.schedule_once(lambda dt, pr=pr: self._on_pr(pr))
+
+    def _on_pr(self, pr: 'PaymentRequest'):
         if not self.wallet:
             self.show_error(_('No wallet loaded.'))
             return
         if pr.verify(self.wallet.contacts):
-            key = pr.get_id()
-            invoice = self.wallet.get_invoice(key)  # FIXME wrong key...
+            invoice = Invoice.from_bip70_payreq(pr, height=0)
             if invoice and self.wallet.get_invoice_status(invoice) == PR_PAID:
                 self.show_error("invoice already paid")
                 self.send_screen.do_clear()
@@ -462,22 +476,15 @@ class ElectrumWindow(App, Logger):
             self.send_screen.do_clear()
 
     def on_qr(self, data: str):
-        from electrum.bitcoin import is_address
+        self.on_data_input(data)
+
+    def on_data_input(self, data: str) -> None:
+        """on_qr / on_paste shared logic"""
         data = data.strip()
-        if is_address(data):
-            self.set_URI(data)
-            return
-        if data.lower().startswith(BITCOIN_BIP21_URI_SCHEME + ':'):
-            self.set_URI(data)
-            return
         if data.lower().startswith('channel_backup:'):
             self.import_channel_backup(data)
             return
-        bolt11_invoice = maybe_extract_bolt11_invoice(data)
-        if bolt11_invoice is not None:
-            self.set_ln_invoice(bolt11_invoice)
-            return
-        # try to decode transaction
+        # try to decode as transaction
         from electrum.transaction import tx_from_any
         try:
             tx = tx_from_any(data)
@@ -486,8 +493,8 @@ class ElectrumWindow(App, Logger):
         if tx:
             self.tx_dialog(tx)
             return
-        # show error
-        self.show_error("Unable to decode QR data")
+        # try to decode as URI/address
+        self.set_URI(data)
 
     def update_tab(self, name):
         s = getattr(self, name + '_screen', None)
@@ -505,17 +512,17 @@ class ElectrumWindow(App, Logger):
         tab = self.tabs.ids[name + '_tab']
         panel.switch_to(tab)
 
-    def show_request(self, is_lightning, key):
+    def show_request(self, key):
         from .uix.dialogs.request_dialog import RequestDialog
         self.request_popup = RequestDialog('Request', key)
         self.request_popup.open()
 
-    def show_invoice(self, is_lightning, key):
+    def show_invoice(self, key):
         from .uix.dialogs.invoice_dialog import InvoiceDialog
         invoice = self.wallet.get_invoice(key)
         if not invoice:
             return
-        data = invoice.invoice if is_lightning else key
+        data = invoice.lightning_invoice if invoice.is_lightning() else key
         self.invoice_popup = InvoiceDialog('Invoice', data, key)
         self.invoice_popup.open()
 
@@ -637,25 +644,7 @@ class ElectrumWindow(App, Logger):
             mactivity = PythonActivity.mActivity
             self.on_new_intent(mactivity.getIntent())
             activity.bind(on_new_intent=self.on_new_intent)
-        # connect callbacks
-        if self.network:
-            interests = ['wallet_updated', 'network_updated', 'blockchain_updated',
-                         'status', 'new_transaction', 'verified']
-            util.register_callback(self.on_network_event, interests)
-            util.register_callback(self.on_fee, ['fee'])
-            util.register_callback(self.on_fee_histogram, ['fee_histogram'])
-            util.register_callback(self.on_quotes, ['on_quotes'])
-            util.register_callback(self.on_history, ['on_history'])
-            util.register_callback(self.on_channels, ['channels_updated'])
-            util.register_callback(self.on_channel, ['channel'])
-            util.register_callback(self.on_invoice_status, ['invoice_status'])
-            util.register_callback(self.on_request_status, ['request_status'])
-            util.register_callback(self.on_payment_failed, ['payment_failed'])
-            util.register_callback(self.on_payment_succeeded, ['payment_succeeded'])
-            util.register_callback(self.on_channel_db, ['channel_db'])
-            util.register_callback(self.set_num_peers, ['gossip_peers'])
-            util.register_callback(self.set_unknown_channels, ['unknown_channels'])
-        
+        self.register_callbacks()
         if self.network and self.electrum_config.get('auto_connect') is None:
             self.popup_dialog("first_screen")
             # load_wallet_on_start will be called later, after initial network setup is completed
@@ -667,14 +656,17 @@ class ElectrumWindow(App, Logger):
             if uri:
                 self.set_URI(uri)
 
-    def on_channel_db(self, event, num_nodes, num_channels, num_policies):
+    @event_listener
+    def on_event_channel_db(self, num_nodes, num_channels, num_policies):
         self.lightning_gossip_num_nodes = num_nodes
         self.lightning_gossip_num_channels = num_channels
 
-    def set_num_peers(self, event, num_peers):
+    @event_listener
+    def on_event_gossip_peers(self, num_peers):
         self.lightning_gossip_num_peers = num_peers
 
-    def set_unknown_channels(self, event, unknown):
+    @event_listener
+    def on_event_unknown_channels(self, unknown):
         self.lightning_gossip_num_queries = unknown
 
     def get_wallet_path(self):
@@ -686,7 +678,8 @@ class ElectrumWindow(App, Logger):
     def on_wizard_success(self, storage, db, password):
         self.password = password
         if self.electrum_config.get('single_password'):
-            self._use_single_password = update_password_for_directory(self.electrum_config, password, password)
+            self._use_single_password = self.daemon.update_password_for_directory(
+                old_password=password, new_password=password)
         self.logger.info(f'use single password: {self._use_single_password}')
         wallet = Wallet(db, storage, config=self.electrum_config)
         wallet.start_network(self.daemon.network)
@@ -781,11 +774,40 @@ class ElectrumWindow(App, Logger):
             self._channels_dialog = LightningChannelsDialog(self)
         self._channels_dialog.open()
 
-    def on_channel(self, evt, wallet, chan):
+    def delete_ln_gossip_dialog(self):
+        def delete_gossip(b: bool):
+            if not b:
+                return
+            if self.network:
+                self.network.run_from_another_thread(
+                    self.network.stop_gossip(full_shutdown=True))
+
+            os.unlink(gossip_db_file)
+            self.show_error(_("Local gossip database deleted."))
+            self.network.start_gossip()
+
+        if self.network is None or self.network.channel_db is None:
+            return  # TODO show msg to user, or the button should be disabled instead
+        gossip_db_file = self.network.channel_db.get_file_path(self.electrum_config)
+        try:
+            size_mb = os.path.getsize(gossip_db_file) / (1024**2)
+        except OSError:
+            self.logger.exception("Cannot get file size.")
+            return
+        d = Question(
+            _('Do you want to delete the local gossip database?') + '\n' +
+            '(' + _('file size') + f': {size_mb:.2f} MiB)\n' +
+            _('It will be automatically re-downloaded after, unless you disable the gossip.'),
+            delete_gossip)
+        d.open()
+
+    @event_listener
+    def on_event_channel(self, wallet, chan):
         if self._channels_dialog:
             Clock.schedule_once(lambda dt: self._channels_dialog.update())
 
-    def on_channels(self, evt, wallet):
+    @event_listener
+    def on_event_channels(self, wallet):
         if self._channels_dialog:
             Clock.schedule_once(lambda dt: self._channels_dialog.update())
 
@@ -867,23 +889,32 @@ class ElectrumWindow(App, Logger):
         self.proxy_config = net_params.proxy or {}
         self.update_proxy_str(self.proxy_config)
 
-    def on_network_event(self, event, *args):
-        self.logger.info('network event: '+ event)
-        if event == 'network_updated':
-            self._trigger_update_interfaces()
-            self._trigger_update_status()
-        elif event == 'wallet_updated':
-            self._trigger_update_wallet()
-            self._trigger_update_status()
-        elif event == 'blockchain_updated':
-            # to update number of confirmations in history
-            self._trigger_update_wallet()
-        elif event == 'status':
-            self._trigger_update_status()
-        elif event == 'new_transaction':
-            self._trigger_update_wallet()
-        elif event == 'verified':
-            self._trigger_update_wallet()
+    @event_listener
+    def on_event_network_updated(self):
+        self._trigger_update_interfaces()
+        self._trigger_update_status()
+
+    @event_listener
+    def on_event_wallet_updated(self, *args):
+        self._trigger_update_wallet()
+        self._trigger_update_status()
+
+    @event_listener
+    def on_event_blockchain_updated(self, *args):
+        # to update number of confirmations in history
+        self._trigger_update_wallet()
+
+    @event_listener
+    def on_event_status(self, *args):
+        self._trigger_update_status()
+
+    @event_listener
+    def on_event_new_transaction(self, *args):
+        self._trigger_update_wallet()
+
+    @event_listener
+    def on_event_verified(self, *args):
+        self._trigger_update_wallet()
 
     @profiler
     def load_wallet(self, wallet: 'Abstract_Wallet'):
@@ -927,8 +958,8 @@ class ElectrumWindow(App, Logger):
             self.num_blocks = self.network.get_local_height()
             server_height = self.network.get_server_height()
             server_lag = self.num_blocks - server_height
-            if not self.wallet.up_to_date or server_height == 0:
-                num_sent, num_answered = self.wallet.get_history_sync_state_details()
+            if not self.wallet.is_up_to_date() or server_height == 0:
+                num_sent, num_answered = self.wallet.adb.get_history_sync_state_details()
                 status = ("{} [size=18dp]({}/{})[/size]"
                           .format(_("Synchronizing..."), num_answered, num_sent))
             elif server_lag > 1:
@@ -951,7 +982,7 @@ class ElectrumWindow(App, Logger):
     def update_wallet_synchronizing_progress(self, *dt):
         if not self.wallet:
             return
-        if not self.wallet.up_to_date:
+        if not self.wallet.is_up_to_date():
             self._trigger_update_status()
 
     def get_max_amount(self):
@@ -1010,7 +1041,7 @@ class ElectrumWindow(App, Logger):
     #@profiler
     def update_wallet(self, *dt):
         self._trigger_update_status()
-        if self.wallet and (self.wallet.up_to_date or not self.network or not self.network.is_connected()):
+        if self.wallet and (self.wallet.is_up_to_date() or not self.network or not self.network.is_connected()):
             self.update_tabs()
 
     def notify(self, message):
@@ -1059,6 +1090,17 @@ class ElectrumWindow(App, Logger):
             return
         self.qr_dialog(label.name, label.data, show_text_with_qr)
 
+    def scheduled_in_gui_thread(func):
+        """Decorator to ensure that func runs in the GUI thread.
+        Note: the return value is swallowed!
+        """
+        def wrapper(self: 'ElectrumWindow', *args, **kwargs):
+            if threading.current_thread() == self.gui_thread:
+                func(self, *args, **kwargs)
+            else:
+                Clock.schedule_once(lambda dt: func(self, *args, **kwargs))
+        return wrapper
+
     def show_error(self, error, width='200dp', pos=None, arrow_pos=None,
                    exit=False, icon=f'atlas://{KIVY_GUI_PATH}/theming/atlas/light/error', duration=0,
                    modal=False):
@@ -1076,6 +1118,7 @@ class ElectrumWindow(App, Logger):
             duration=duration, modal=modal, exit=exit, pos=pos,
             arrow_pos=arrow_pos)
 
+    @scheduled_in_gui_thread
     def show_info_bubble(self, text=_('Hello World'), pos=None, duration=0,
                          arrow_pos='bottom_mid', width=None, icon='', modal=False, exit=False):
         '''Method to show an Information Bubble
@@ -1135,7 +1178,7 @@ class ElectrumWindow(App, Logger):
     def show_transaction(self, txid):
         tx = self.wallet.db.get_transaction(txid)
         if not tx and self.wallet.lnworker:
-            tx = self.wallet.lnworker.lnwatcher.db.get_transaction(txid)
+            tx = self.wallet.adb.get_transaction(txid)
         if tx:
             self.tx_dialog(tx)
         else:
@@ -1228,7 +1271,8 @@ class ElectrumWindow(App, Logger):
         target, tooltip, dyn = self.electrum_config.get_fee_target()
         self.fee_status = target
 
-    def on_fee(self, event, *arg):
+    @event_listener
+    def on_event_fee(self, *arg):
         self.set_fee_status()
 
     def protected(self, msg, f, args):
@@ -1302,10 +1346,7 @@ class ElectrumWindow(App, Logger):
             # called if old_password works on self.wallet
             self.password = new_password
             if self._use_single_password:
-                path = self.wallet.storage.path
-                self.stop_wallet()
-                update_password_for_directory(self.electrum_config, old_password, new_password)
-                self.load_wallet_by_name(path)
+                self.daemon.update_password_for_directory(old_password=old_password, new_password=new_password)
                 msg = _("Password updated successfully")
             else:
                 self.wallet.update_password(old_password, new_password)
@@ -1411,6 +1452,14 @@ class ElectrumWindow(App, Logger):
         self.protected(_("Decrypt your private key?"), show_private_key, (addr, pk_label))
 
     def import_channel_backup(self, encrypted):
+        if not self.wallet.has_lightning():
+            msg = _('Cannot import channel backup.')
+            if self.wallet.can_have_lightning():
+                msg += ' ' + _('Lightning is not enabled.')
+            else:
+                msg += ' ' + _('Lightning is not available for this wallet.')
+            self.show_error(msg)
+            return
         d = Question(_('Import Channel Backup?'), lambda b: self._import_channel_backup(b, encrypted))
         d.open()
 
@@ -1461,7 +1510,7 @@ class ElectrumWindow(App, Logger):
             else:
                 msg = _(
                     "Warning: this wallet type does not support channel recovery from seed. "
-                    "You will need to backup your wallet everytime you create a new wallet. "
+                    "You will need to backup your wallet everytime you create a new channel. "
                     "Create lightning keys?")
             d = Question(msg, self._enable_lightning, title=_('Enable Lightning?'))
             d.open()
