@@ -28,13 +28,17 @@ import signal
 import sys
 import traceback
 import threading
-from typing import Optional, TYPE_CHECKING, List
+from typing import Optional, TYPE_CHECKING, List, Sequence
 
+from electrum import GuiImportError
 
 try:
     import PyQt5
-except Exception:
-    sys.exit("Error: Could not import PyQt5 on Linux systems, you may try 'sudo apt-get install python3-pyqt5'")
+    import PyQt5.QtGui
+except Exception as e:
+    raise GuiImportError(
+        "Error: Could not import PyQt5 on Linux systems, "
+        "you may try 'sudo apt-get install python3-pyqt5'") from e
 
 from PyQt5.QtGui import QGuiApplication
 from PyQt5.QtWidgets import (QApplication, QSystemTrayIcon, QWidget, QMenu,
@@ -42,6 +46,14 @@ from PyQt5.QtWidgets import (QApplication, QSystemTrayIcon, QWidget, QMenu,
 from PyQt5.QtCore import QObject, pyqtSignal, QTimer, Qt
 import PyQt5.QtCore as QtCore
 
+try:
+    # Preload QtMultimedia at app start, if available.
+    # We use QtMultimedia on some platforms for camera-handling, and
+    # lazy-loading it later led to some crashes. Maybe due to bugs in PyQt5. (see #7725)
+    from PyQt5.QtMultimedia import QCameraInfo; del QCameraInfo
+except ImportError as e:
+    pass  # failure is ok; it is an optional dependency.
+
 from electrum.i18n import _, set_language
 from electrum.plugin import run_hook
 from electrum.base_wizard import GoBack
@@ -68,14 +80,14 @@ if TYPE_CHECKING:
 
 
 class OpenFileEventFilter(QObject):
-    def __init__(self, windows):
+    def __init__(self, windows: Sequence[ElectrumWindow]):
         self.windows = windows
         super(OpenFileEventFilter, self).__init__()
 
     def eventFilter(self, obj, event):
         if event.type() == QtCore.QEvent.FileOpen:
             if len(self.windows) >= 1:
-                self.windows[0].pay_to_URI(event.url().toString())
+                self.windows[0].handle_payment_identifier(event.url().toString())
                 return True
         return False
 
@@ -83,11 +95,13 @@ class OpenFileEventFilter(QObject):
 class QElectrumApplication(QApplication):
     new_window_signal = pyqtSignal(str, object)
     quit_signal = pyqtSignal()
+    refresh_tabs_signal = pyqtSignal()
+    refresh_amount_edits_signal = pyqtSignal()
+    update_status_signal = pyqtSignal()
+    update_fiat_signal = pyqtSignal()
+    alias_received_signal = pyqtSignal()
 
 
-class QNetworkUpdatedSignalObject(QObject):
-    network_updated_signal = pyqtSignal(str, object)
-
 
 class ElectrumGui(BaseElectrumGui, Logger):
 
@@ -125,7 +139,6 @@ class ElectrumGui(BaseElectrumGui, Logger):
         self.network_dialog = None
         self.lightning_dialog = None
         self.watchtower_dialog = None
-        self.network_updated_signal_obj = QNetworkUpdatedSignalObject()
         self._num_wizards_in_progress = 0
         self._num_wizards_lock = threading.Lock()
         self.dark_icon = self.config.get("dark_icon", False)
@@ -234,7 +247,6 @@ class ElectrumGui(BaseElectrumGui, Logger):
             self.network_dialog.close()
             self.network_dialog.clean_up()
             self.network_dialog = None
-        self.network_updated_signal_obj = None
         if self.lightning_dialog:
             self.lightning_dialog.close()
             self.lightning_dialog = None
@@ -281,14 +293,13 @@ class ElectrumGui(BaseElectrumGui, Logger):
 
     def show_network_dialog(self):
         if self.network_dialog:
-            self.network_dialog.on_update()
+            self.network_dialog.on_event_network_updated()
             self.network_dialog.show()
             self.network_dialog.raise_()
             return
         self.network_dialog = NetworkDialog(
             network=self.daemon.network,
-            config=self.config,
-            network_updated_signal_obj=self.network_updated_signal_obj)
+            config=self.config)
         self.network_dialog.show()
 
     def _create_window_for_wallet(self, wallet):
@@ -312,21 +323,33 @@ class ElectrumGui(BaseElectrumGui, Logger):
         return wrapper
 
     @count_wizards_in_progress
-    def start_new_window(self, path, uri, *, app_is_starting=False) -> Optional[ElectrumWindow]:
+    def start_new_window(
+            self,
+            path,
+            uri: Optional[str],
+            *,
+            app_is_starting: bool = False,
+            force_wizard: bool = False,
+    ) -> Optional[ElectrumWindow]:
         '''Raises the window for the wallet if it is open.  Otherwise
         opens the wallet and creates a new window for it'''
         wallet = None
-        try:
-            wallet = self.daemon.load_wallet(path, None)
-        except Exception as e:
-            self.logger.exception('')
-            custom_message_box(icon=QMessageBox.Warning,
-                               parent=None,
-                               title=_('Error'),
-                               text=_('Cannot load wallet') + ' (1):\n' + repr(e))
-            # if app is starting, still let wizard appear
-            if not app_is_starting:
-                return
+        # Try to open with daemon first. If this succeeds, there won't be a wizard at all
+        # (the wallet main window will appear directly).
+        if not force_wizard:
+            try:
+                wallet = self.daemon.load_wallet(path, None)
+            except Exception as e:
+                self.logger.exception('')
+                custom_message_box(icon=QMessageBox.Warning,
+                                   parent=None,
+                                   title=_('Error'),
+                                   text=_('Cannot load wallet') + ' (1):\n' + repr(e))
+                # if app is starting, still let wizard appear
+                if not app_is_starting:
+                    return
+        # Open a wizard window. This lets the user e.g. enter a password, or select
+        # a different wallet.
         try:
             if not wallet:
                 wallet = self._start_wizard_to_select_or_create_wallet(path)
@@ -345,12 +368,22 @@ class ElectrumGui(BaseElectrumGui, Logger):
                                title=_('Error'),
                                text=_('Cannot load wallet') + '(2) :\n' + repr(e))
             if app_is_starting:
-                wallet_dir = os.path.dirname(path)
-                path = os.path.join(wallet_dir, get_new_wallet_name(wallet_dir))
-                self.start_new_window(path, uri)
+                # If we raise in this context, there are no more fallbacks, we will shut down.
+                # Worst case scenario, we might have gotten here without user interaction,
+                # in which case, if we raise now without user interaction, the same sequence of
+                # events is likely to repeat when the user restarts the process.
+                # So we play it safe: clear path, clear uri, force a wizard to appear.
+                try:
+                    wallet_dir = os.path.dirname(path)
+                    filename = get_new_wallet_name(wallet_dir)
+                except OSError:
+                    path = self.config.get_fallback_wallet_path()
+                else:
+                    path = os.path.join(wallet_dir, filename)
+                self.start_new_window(path, uri=None, force_wizard=True)
             return
         if uri:
-            window.pay_to_URI(uri)
+            window.handle_payment_identifier(uri)
         window.bring_to_top()
         window.setWindowState(window.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
 
@@ -438,3 +471,13 @@ class ElectrumGui(BaseElectrumGui, Logger):
     def stop(self):
         self.logger.info('closing GUI')
         self.app.quit_signal.emit()
+
+    @classmethod
+    def version_info(cls):
+        ret = {
+            "qt.version": QtCore.QT_VERSION_STR,
+            "pyqt.version": QtCore.PYQT_VERSION_STR,
+        }
+        if hasattr(PyQt5, "__path__"):
+            ret["pyqt.path"] = ", ".join(PyQt5.__path__ or [])
+        return ret
