@@ -32,7 +32,6 @@ from PyQt5.QtCore import Qt, QItemSelectionModel, QModelIndex
 
 from electrum.i18n import _
 from electrum.util import format_time
-from electrum.invoices import PR_TYPE_ONCHAIN, PR_TYPE_LN, LNInvoice, OnchainInvoice
 from electrum.plugin import run_hook
 from electrum.invoices import Invoice
 
@@ -40,6 +39,7 @@ from .util import MyTreeView, pr_icons, read_QIcon, webopen, MySortModel
 
 if TYPE_CHECKING:
     from .main_window import ElectrumWindow
+    from .receive_tab import ReceiveTab
 
 
 ROLE_REQUEST_TYPE = Qt.UserRole
@@ -64,10 +64,12 @@ class RequestList(MyTreeView):
     }
     filter_columns = [Columns.DATE, Columns.DESCRIPTION, Columns.AMOUNT]
 
-    def __init__(self, parent: 'ElectrumWindow'):
-        super().__init__(parent, self.create_menu,
+    def __init__(self, receive_tab: 'ReceiveTab'):
+        window = receive_tab.window
+        super().__init__(window, self.create_menu,
                          stretch_column=self.Columns.DESCRIPTION)
-        self.wallet = self.parent.wallet
+        self.wallet = window.wallet
+        self.receive_tab = receive_tab
         self.std_model = QStandardItemModel(self)
         self.proxy = MySortModel(self, sort_role=ROLE_SORT_ORDER)
         self.proxy.setSourceModel(self.std_model)
@@ -75,9 +77,8 @@ class RequestList(MyTreeView):
         self.setSortingEnabled(True)
         self.selectionModel().currentRowChanged.connect(self.item_changed)
         self.setSelectionMode(QAbstractItemView.ExtendedSelection)
-        self.update()
 
-    def select_key(self, key):
+    def set_current_key(self, key):
         for i in range(self.model().rowCount()):
             item = self.model().index(i, self.Columns.DATE)
             row_key = item.data(ROLE_KEY)
@@ -85,10 +86,12 @@ class RequestList(MyTreeView):
                 self.selectionModel().setCurrentIndex(item, QItemSelectionModel.SelectCurrent | QItemSelectionModel.Rows)
                 break
 
+    def get_current_key(self):
+        return self.get_role_data_for_current_item(col=self.Columns.DATE, role=ROLE_KEY)
+
     def item_changed(self, idx: Optional[QModelIndex]):
         if idx is None:
-            self.parent.receive_payreq_e.setText('')
-            self.parent.receive_address_e.setText('')
+            self.receive_tab.update_current_request()
             return
         if not idx.isValid():
             return
@@ -98,73 +101,59 @@ class RequestList(MyTreeView):
         req = self.wallet.get_request(key)
         if req is None:
             self.update()
-            return
-        if req.is_lightning():
-            self.parent.receive_payreq_e.setText(req.invoice)  # TODO maybe prepend "lightning:" ??
-            self.parent.receive_address_e.setText(req.invoice)
-        else:
-            self.parent.receive_payreq_e.setText(self.parent.wallet.get_request_URI(req))
-            self.parent.receive_address_e.setText(req.get_address())
-        self.parent.receive_payreq_e.repaint()  # macOS hack (similar to #4777)
-        self.parent.receive_address_e.repaint()  # macOS hack (similar to #4777)
+        self.receive_tab.update_current_request()
 
     def clearSelection(self):
         super().clearSelection()
         self.selectionModel().clearCurrentIndex()
 
     def refresh_row(self, key, row):
+        assert row is not None
         model = self.std_model
         request = self.wallet.get_request(key)
         if request is None:
             return
         status_item = model.item(row, self.Columns.STATUS)
-        status = self.parent.wallet.get_request_status(key)
+        status = self.wallet.get_request_status(key)
         status_str = request.get_status_str(status)
         status_item.setText(status_str)
         status_item.setIcon(read_QIcon(pr_icons.get(status)))
 
     def update(self):
+        current_key = self.get_current_key()
         # not calling maybe_defer_update() as it interferes with conditional-visibility
-        self.parent.update_receive_address_styling()
         self.proxy.setDynamicSortFilter(False)  # temp. disable re-sorting after every change
         self.std_model.clear()
         self.update_headers(self.__class__.headers)
         for req in self.wallet.get_unpaid_requests():
             key = self.wallet.get_key_for_receive_request(req)
-            status = self.parent.wallet.get_request_status(key)
+            status = self.wallet.get_request_status(key)
             status_str = req.get_status_str(status)
-            request_type = req.type
-            timestamp = req.time
+            timestamp = req.get_time()
             amount = req.get_amount_sat()
-            message = req.message
+            message = req.get_message()
             date = format_time(timestamp)
             amount_str = self.parent.format_amount(amount) if amount else ""
             labels = [date, message, amount_str, status_str]
-            if req.is_lightning():
-                icon = read_QIcon("lightning.png")
-                tooltip = 'lightning request'
-            else:
-                icon = read_QIcon("bitcoin.png")
-                tooltip = 'onchain request'
             items = [QStandardItem(e) for e in labels]
             self.set_editability(items)
-            items[self.Columns.DATE].setData(request_type, ROLE_REQUEST_TYPE)
+            #items[self.Columns.DATE].setData(request_type, ROLE_REQUEST_TYPE)
             items[self.Columns.DATE].setData(key, ROLE_KEY)
             items[self.Columns.DATE].setData(timestamp, ROLE_SORT_ORDER)
-            items[self.Columns.DATE].setIcon(icon)
             items[self.Columns.STATUS].setIcon(read_QIcon(pr_icons.get(status)))
-            items[self.Columns.DATE].setToolTip(tooltip)
             self.std_model.insertRow(self.std_model.rowCount(), items)
         self.filter()
         self.proxy.setDynamicSortFilter(True)
         # sort requests by date
         self.sortByColumn(self.Columns.DATE, Qt.DescendingOrder)
         self.hide_if_empty()
+        if current_key is not None:
+            self.set_current_key(current_key)
 
     def hide_if_empty(self):
         b = self.std_model.rowCount() > 0
         self.setVisible(b)
-        self.parent.receive_requests_label.setVisible(b)
+        self.receive_tab.receive_requests_label.setVisible(b)
         if not b:
             # list got hidden, so selected item should also be cleared:
             self.item_changed(None)
@@ -174,7 +163,7 @@ class RequestList(MyTreeView):
         if len(items)>1:
             keys = [item.data(ROLE_KEY)  for item in items]
             menu = QMenu(self)
-            menu.addAction(_("Delete requests"), lambda: self.parent.delete_requests(keys))
+            menu.addAction(_("Delete requests"), lambda: self.delete_requests(keys))
             menu.exec_(self.viewport().mapToGlobal(position))
             return
         idx = self.indexAt(position)
@@ -188,15 +177,21 @@ class RequestList(MyTreeView):
             self.update()
             return
         menu = QMenu(self)
-        self.add_copy_menu(menu, idx)
-        if req.is_lightning():
-            menu.addAction(_("Copy Request"), lambda: self.parent.do_copy(req.invoice, title='Lightning Request'))
-        else:
-            URI = self.wallet.get_request_URI(req)
-            menu.addAction(_("Copy Request"), lambda: self.parent.do_copy(URI, title='Bitcoin URI'))
+        if req.get_address():
             menu.addAction(_("Copy Address"), lambda: self.parent.do_copy(req.get_address(), title='Bitcoin Address'))
+            URI = self.wallet.get_request_URI(req)
+            menu.addAction(_("Copy URI"), lambda: self.parent.do_copy(URI, title='Bitcoin URI'))
+        if req.is_lightning():
+            menu.addAction(_("Copy Lightning Request"), lambda: self.parent.do_copy(req.lightning_invoice, title='Lightning Request'))
+        self.add_copy_menu(menu, idx)
         #if 'view_url' in req:
         #    menu.addAction(_("View in web browser"), lambda: webopen(req['view_url']))
-        menu.addAction(_("Delete"), lambda: self.parent.delete_requests([key]))
+        menu.addAction(_("Delete"), lambda: self.delete_requests([key]))
         run_hook('receive_list_menu', self.parent, menu, key)
         menu.exec_(self.viewport().mapToGlobal(position))
+
+    def delete_requests(self, keys):
+        for key in keys:
+            self.wallet.delete_request(key)
+            self.delete_item(key)
+        self.receive_tab.do_clear()
