@@ -2,12 +2,14 @@ from decimal import Decimal
 import getpass
 import datetime
 import logging
+from typing import Optional
 
 from electrum.gui import BaseElectrumGui
 from electrum import util
 from electrum import WalletStorage, Wallet
+from electrum.wallet import Abstract_Wallet
 from electrum.wallet_db import WalletDB
-from electrum.util import format_satoshis
+from electrum.util import format_satoshis, EventListener, event_listener
 from electrum.bitcoin import is_address, COIN
 from electrum.transaction import PartialTxOutput
 from electrum.network import TxBroadcastError, BestEffortRequestFailed
@@ -18,13 +20,13 @@ _ = lambda x:x  # i18n
 # written by rofl0r, with some bits stolen from the text gui (ncurses)
 
 
-class ElectrumGui(BaseElectrumGui):
+class ElectrumGui(BaseElectrumGui, EventListener):
 
     def __init__(self, *, config, daemon, plugins):
         BaseElectrumGui.__init__(self, config=config, daemon=daemon, plugins=plugins)
         self.network = daemon.network
         storage = WalletStorage(config.get_wallet_path())
-        if not storage.file_exists:
+        if not storage.file_exists():
             print("Wallet not found. try 'electrum create'")
             exit()
         if storage.is_encrypted():
@@ -41,11 +43,11 @@ class ElectrumGui(BaseElectrumGui):
         self.str_amount = ""
         self.str_fee = ""
 
-        self.wallet = Wallet(db, storage, config=config)
+        self.wallet = Wallet(db, storage, config=config)  # type: Optional[Abstract_Wallet]
         self.wallet.start_network(self.network)
         self.contacts = self.wallet.contacts
 
-        util.register_callback(self.on_network, ['wallet_updated', 'network_updated', 'banner'])
+        self.register_callbacks()
         self.commands = [_("[h] - displays this help text"), \
                          _("[i] - display transaction history"), \
                          _("[o] - enter payment order"), \
@@ -57,11 +59,17 @@ class ElectrumGui(BaseElectrumGui):
                          _("[q] - quit")]
         self.num_commands = len(self.commands)
 
-    def on_network(self, event, *args):
-        if event in ['wallet_updated', 'network_updated']:
-            self.updated()
-        elif event == 'banner':
-            self.print_banner()
+    @event_listener
+    def on_event_wallet_updated(self, wallet):
+        self.updated()
+
+    @event_listener
+    def on_event_network_updated(self):
+        self.updated()
+
+    @event_listener
+    def on_event_banner(self):
+        self.print_banner()
 
     def main_command(self):
         self.print_balance()
@@ -95,8 +103,8 @@ class ElectrumGui(BaseElectrumGui):
         format_str = "%"+"%d"%width[0]+"s"+"%"+"%d"%(width[1]+delta)+"s"+"%" \
         + "%d"%(width[2]+delta)+"s"+"%"+"%d"%(width[3]+delta)+"s"
         messages = []
-
-        for hist_item in reversed(self.wallet.get_history()):
+        domain = self.wallet.get_addresses()
+        for hist_item in reversed(self.wallet.adb.get_history(domain)):
             if hist_item.tx_mined_status.conf:
                 timestamp = hist_item.tx_mined_status.timestamp
                 try:
@@ -118,7 +126,7 @@ class ElectrumGui(BaseElectrumGui):
 
     def get_balance(self):
         if self.wallet.network.is_connected():
-            if not self.wallet.up_to_date:
+            if not self.wallet.is_up_to_date():
                 msg = _("Synchronizing...")
             else:
                 c, u, x =  self.wallet.get_balance()
