@@ -17,12 +17,12 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
-
+import enum
 import os
 from collections import namedtuple, defaultdict
 import binascii
 import json
-from enum import IntEnum
+from enum import IntEnum, Enum
 from typing import (Optional, Dict, List, Tuple, NamedTuple, Set, Callable,
                     Iterable, Sequence, TYPE_CHECKING, Iterator, Union, Mapping)
 import time
@@ -82,11 +82,14 @@ class ChannelState(IntEnum):
     OPEN            = 3  # both parties have sent funding_locked
     SHUTDOWN        = 4  # shutdown has been sent.
     CLOSING         = 5  # closing negotiation done. we have a fully signed tx.
-    FORCE_CLOSING   = 6  # we force-closed, and closing tx is unconfirmed. Note that if the
+    FORCE_CLOSING   = 6  # *we* force-closed, and closing tx is unconfirmed. Note that if the
                          # remote force-closes then we remain OPEN until it gets mined -
                          # the server could be lying to us with a fake tx.
-    CLOSED          = 7  # closing tx has been mined
-    REDEEMED        = 8  # we can stop watching
+    REQUESTED_FCLOSE = 7   # Chan is open, but we have tried to request the *remote* to force-close
+    WE_ARE_TOXIC     = 8   # Chan is open, but we have lost state and the remote proved this.
+                           # The remote must force-close, it is *not* safe for us to do so.
+    CLOSED           = 9   # closing tx has been mined
+    REDEEMED         = 10  # we can stop watching
 
 
 class PeerState(IntEnum):
@@ -113,12 +116,28 @@ state_transitions = [
     (cs.OPEN,     cs.FORCE_CLOSING),
     (cs.SHUTDOWN, cs.FORCE_CLOSING),
     (cs.CLOSING,  cs.FORCE_CLOSING),
+    (cs.REQUESTED_FCLOSE, cs.FORCE_CLOSING),
+    # we can request a force-close almost any time
+    (cs.OPENING,  cs.REQUESTED_FCLOSE),
+    (cs.FUNDED,   cs.REQUESTED_FCLOSE),
+    (cs.OPEN,     cs.REQUESTED_FCLOSE),
+    (cs.SHUTDOWN, cs.REQUESTED_FCLOSE),
+    (cs.CLOSING,  cs.REQUESTED_FCLOSE),
+    (cs.REQUESTED_FCLOSE,  cs.REQUESTED_FCLOSE),
     # we can get force closed almost any time
     (cs.OPENING,  cs.CLOSED),
     (cs.FUNDED,   cs.CLOSED),
     (cs.OPEN,     cs.CLOSED),
     (cs.SHUTDOWN, cs.CLOSED),
     (cs.CLOSING,  cs.CLOSED),
+    (cs.REQUESTED_FCLOSE, cs.CLOSED),
+    (cs.WE_ARE_TOXIC,          cs.CLOSED),
+    # during channel_reestablish, we might realise we have lost state
+    (cs.OPENING,  cs.WE_ARE_TOXIC),
+    (cs.FUNDED,   cs.WE_ARE_TOXIC),
+    (cs.OPEN,     cs.WE_ARE_TOXIC),
+    (cs.SHUTDOWN, cs.WE_ARE_TOXIC),
+    (cs.REQUESTED_FCLOSE, cs.WE_ARE_TOXIC),
     #
     (cs.FORCE_CLOSING, cs.FORCE_CLOSING),  # allow multiple attempts
     (cs.FORCE_CLOSING, cs.CLOSED),
@@ -130,6 +149,12 @@ state_transitions = [
 del cs  # delete as name is ambiguous without context
 
 
+class ChanCloseOption(Enum):
+    COOP_CLOSE = enum.auto()
+    LOCAL_FCLOSE = enum.auto()
+    REQUEST_REMOTE_FCLOSE = enum.auto()
+
+
 class RevokeAndAck(NamedTuple):
     per_commitment_secret: bytes
     next_per_commitment_point: bytes
@@ -156,6 +181,7 @@ class AbstractChannel(Logger, ABC):
     lnworker: Optional['LNWallet']
     _fallback_sweep_address: str
     channel_id: bytes
+    short_channel_id: Optional[ShortChannelID] = None
     funding_outpoint: Outpoint
     node_id: bytes  # note that it might not be the full 33 bytes; for OCB it is only the prefix
     _state: ChannelState
@@ -180,7 +206,7 @@ class AbstractChannel(Logger, ABC):
         old_state = self._state
         if not force and (old_state, state) not in state_transitions:
             raise Exception(f"Transition not allowed: {old_state.name} -> {state.name}")
-        self.logger.debug(f'({self.get_id_for_log()}) Setting channel state: {old_state.name} -> {state.name}')
+        self.logger.debug(f'Setting channel state: {old_state.name} -> {state.name}')
         self._state = state
         self.storage['state'] = self._state.name
         if self.lnworker:
@@ -195,9 +221,6 @@ class AbstractChannel(Logger, ABC):
     def is_open(self):
         return self.get_state() == ChannelState.OPEN
 
-    def is_closing(self):
-        return ChannelState.SHUTDOWN <= self.get_state() <= ChannelState.FORCE_CLOSING
-
     def is_closed(self):
         # the closing txid has been saved
         return self.get_state() >= ChannelState.CLOSING
@@ -205,6 +228,10 @@ class AbstractChannel(Logger, ABC):
     def is_redeemed(self):
         return self.get_state() == ChannelState.REDEEMED
 
+    @abstractmethod
+    def get_close_options(self) -> Sequence[ChanCloseOption]:
+        pass
+
     def save_funding_height(self, *, txid: str, height: int, timestamp: Optional[int]) -> None:
         self.storage['funding_height'] = txid, height, timestamp
 
@@ -255,13 +282,16 @@ class AbstractChannel(Logger, ABC):
         if funding_height.height == TX_HEIGHT_LOCAL:
             self.update_unfunded_state()
         elif closing_height.height == TX_HEIGHT_LOCAL:
-            self.update_funded_state(funding_txid=funding_txid, funding_height=funding_height)
+            self.update_funded_state(
+                funding_txid=funding_txid,
+                funding_height=funding_height)
         else:
-            self.update_closed_state(funding_txid=funding_txid,
-                                     funding_height=funding_height,
-                                     closing_txid=closing_txid,
-                                     closing_height=closing_height,
-                                     keep_watching=keep_watching)
+            self.update_closed_state(
+                funding_txid=funding_txid,
+                funding_height=funding_height,
+                closing_txid=closing_txid,
+                closing_height=closing_height,
+                keep_watching=keep_watching)
 
     def update_unfunded_state(self):
         self.delete_funding_height()
@@ -280,14 +310,14 @@ class AbstractChannel(Logger, ABC):
                     if spender_txid is None:
                         continue
                     if spender_txid != self.funding_outpoint.txid:
-                        tx_mined_height = self.lnworker.wallet.get_tx_height(spender_txid)
+                        tx_mined_height = self.lnworker.wallet.adb.get_tx_height(spender_txid)
                         if tx_mined_height.conf > lnutil.REDEEM_AFTER_DOUBLE_SPENT_DELAY:
                             self.logger.info(f'channel is double spent {inputs}')
                             self.set_state(ChannelState.REDEEMED)
                             break
             else:
                 now = int(time.time())
-                if now - self.storage.get('init_timestamp', 0) > CHANNEL_OPENING_TIMEOUT:
+                if self.lnworker and (now - self.storage.get('init_timestamp', 0) > CHANNEL_OPENING_TIMEOUT):
                     self.lnworker.remove_channel(self.channel_id)
 
     def update_funded_state(self, *, funding_txid: str, funding_height: TxMinedInfo) -> None:
@@ -321,6 +351,9 @@ class AbstractChannel(Logger, ABC):
 
         if self.get_state() == ChannelState.CLOSED and not keep_watching:
             self.set_state(ChannelState.REDEEMED)
+            if self.lnworker and self.is_backup():
+                # auto-remove redeemed backups
+                self.lnworker.remove_channel_backup(self.channel_id)
 
     @property
     def sweep_address(self) -> str:
@@ -407,6 +440,11 @@ class AbstractChannel(Logger, ABC):
         """Returns our node ID."""
         pass
 
+    @abstractmethod
+    def get_capacity(self) -> Optional[int]:
+        """Returns channel capacity in satoshis, or None if unknown."""
+        pass
+
 
 class ChannelBackup(AbstractChannel):
     """
@@ -478,7 +516,11 @@ class ChannelBackup(AbstractChannel):
         return self.is_imported or self.is_redeemed()
 
     def get_capacity(self):
-        return self.lnworker.lnwatcher.get_tx_delta(self.funding_outpoint.txid, self.cb.funding_address)
+        lnwatcher = self.lnworker.lnwatcher
+        if lnwatcher:
+            # fixme: we should probably not call that method here
+            return lnwatcher.adb.get_tx_delta(self.funding_outpoint.txid, self.cb.funding_address)
+        return None
 
     def is_backup(self):
         return True
@@ -538,6 +580,12 @@ class ChannelBackup(AbstractChannel):
             return self.lnworker.node_keypair.pubkey
         raise NotImplementedError(f"unexpected cb type: {type(cb)}")
 
+    def get_close_options(self) -> Sequence[ChanCloseOption]:
+        ret = []
+        if self.get_state() == ChannelState.FUNDED:
+            ret.append(ChanCloseOption.REQUEST_REMOTE_FCLOSE)
+        return ret
+
 
 class Channel(AbstractChannel):
     # note: try to avoid naming ctns/ctxs/etc as "current" and "pending".
@@ -549,9 +597,14 @@ class Channel(AbstractChannel):
     forwarding_fee_base_msat = 1000
     forwarding_fee_proportional_millionths = 1
 
+    def __repr__(self):
+        return "Channel(%s)"%self.get_id_for_log()
+
     def __init__(self, state: 'StoredDict', *, sweep_address=None, name=None, lnworker=None, initial_feerate=None):
         self.name = name
-        Logger.__init__(self)
+        self.channel_id = bfh(state["channel_id"])
+        self.short_channel_id = ShortChannelID.normalize(state["short_channel_id"])
+        Logger.__init__(self)  # should be after short_channel_id is set
         self.lnworker = lnworker
         self._fallback_sweep_address = sweep_address
         self.storage = state
@@ -559,11 +612,9 @@ class Channel(AbstractChannel):
         self.config = {}
         self.config[LOCAL] = state["local_config"]
         self.config[REMOTE] = state["remote_config"]
-        self.channel_id = bfh(state["channel_id"])
         self.constraints = state["constraints"]  # type: ChannelConstraints
         self.funding_outpoint = state["funding_outpoint"]
         self.node_id = bfh(state["node_id"])
-        self.short_channel_id = ShortChannelID.normalize(state["short_channel_id"])
         self.onion_keys = state['onion_keys']  # type: Dict[int, bytes]
         self.data_loss_protect_remote_pcp = state['data_loss_protect_remote_pcp']
         self.hm = HTLCManager(log=state['log'], initial_feerate=initial_feerate)
@@ -601,10 +652,7 @@ class Channel(AbstractChannel):
     def diagnostic_name(self):
         if self.name:
             return str(self.name)
-        try:
-            return f"lnchannel_{bh2u(self.channel_id[-4:])}"
-        except:
-            return super().diagnostic_name()
+        return self.get_id_for_log()
 
     def set_onion_key(self, key: int, value: bytes):
         self.onion_keys[key] = value
@@ -775,7 +823,7 @@ class Channel(AbstractChannel):
 
     def can_send_ctx_updates(self) -> bool:
         """Whether we can send update_fee, update_*_htlc changes to the remote."""
-        if not (self.is_open() or self.is_closing()):
+        if self.get_state() not in (ChannelState.OPEN, ChannelState.SHUTDOWN):
             return False
         if self.peer_state != PeerState.GOOD:
             return False
@@ -784,7 +832,7 @@ class Channel(AbstractChannel):
         return True
 
     def can_send_update_add_htlc(self) -> bool:
-        return self.can_send_ctx_updates() and not self.is_closing()
+        return self.can_send_ctx_updates() and self.is_open()
 
     def is_frozen_for_sending(self) -> bool:
         if self.lnworker and self.lnworker.channel_db is None and not self.lnworker.is_trampoline_peer(self.node_id):
@@ -879,7 +927,9 @@ class Channel(AbstractChannel):
         return True
 
     def should_try_to_reestablish_peer(self) -> bool:
-        return ChannelState.PREOPENING < self._state < ChannelState.CLOSING and self.peer_state == PeerState.DISCONNECTED
+        if self.peer_state != PeerState.DISCONNECTED:
+            return False
+        return ChannelState.PREOPENING < self._state < ChannelState.CLOSING
 
     def get_funding_address(self):
         script = funding_output_script(self.config[LOCAL], self.config[REMOTE])
@@ -1149,6 +1199,9 @@ class Channel(AbstractChannel):
             ctn = self.get_next_ctn(ctx_owner)
         return htlcsum(self.hm.htlcs_by_direction(ctx_owner, direction, ctn).values())
 
+    def has_unsettled_htlcs(self) -> bool:
+        return len(self.hm.htlcs(LOCAL)) + len(self.hm.htlcs(REMOTE)) > 0
+
     def available_to_spend(self, subject: HTLCOwner, *, strict: bool = True) -> int:
         """The usable balance of 'subject' in msat, after taking reserve and fees into
         consideration. Note that fees (and hence the result) fluctuate even without user interaction.
@@ -1490,6 +1543,23 @@ class Channel(AbstractChannel):
         assert tx.is_complete()
         return tx
 
+    def get_close_options(self) -> Sequence[ChanCloseOption]:
+        # This method is used both in the GUI, and in lnpeer.schedule_force_closing
+        # in the latter case, the result does not depend on peer_state
+        ret = []
+        if not self.is_closed() and self.peer_state == PeerState.GOOD:
+            # If there are unsettled HTLCs, althought is possible to cooperatively close,
+            # we choose not to expose that option in the GUI, because it is very likely
+            # that HTLCs will take a long time to settle (submarine swap, or stuck payment),
+            # and the close dialog would be taking a very long time to finish
+            if not self.has_unsettled_htlcs():
+                ret.append(ChanCloseOption.COOP_CLOSE)
+                ret.append(ChanCloseOption.REQUEST_REMOTE_FCLOSE)
+        if not self.is_closed() or self.get_state() == ChannelState.REQUESTED_FCLOSE:
+            ret.append(ChanCloseOption.LOCAL_FCLOSE)
+        assert not (self.get_state() == ChannelState.WE_ARE_TOXIC and ChanCloseOption.LOCAL_FCLOSE in ret), "local force-close unsafe if we are toxic"
+        return ret
+
     def maybe_sweep_revoked_htlc(self, ctx: Transaction, htlc_tx: Transaction) -> Optional[SweepInfo]:
         # look at the output address, check if it matches
         return create_sweeptx_for_their_revoked_htlc(self, ctx, htlc_tx, self.sweep_address)
@@ -1506,8 +1576,8 @@ class Channel(AbstractChannel):
         # to the present, then unilaterally close channel
         recv_htlc_deadline = lnutil.NBLOCK_DEADLINE_BEFORE_EXPIRY_FOR_RECEIVED_HTLCS
         for sub, dir, ctn in ((LOCAL, RECEIVED, self.get_latest_ctn(LOCAL)),
-                              (REMOTE, SENT, self.get_oldest_unrevoked_ctn(LOCAL)),
-                              (REMOTE, SENT, self.get_latest_ctn(LOCAL)),):
+                              (REMOTE, SENT, self.get_oldest_unrevoked_ctn(REMOTE)),
+                              (REMOTE, SENT, self.get_latest_ctn(REMOTE)),):
             for htlc_id, htlc in self.hm.htlcs_by_direction(subject=sub, direction=dir, ctn=ctn).items():
                 if not self.hm.was_htlc_preimage_released(htlc_id=htlc_id, htlc_proposer=REMOTE):
                     continue
@@ -1518,8 +1588,8 @@ class Channel(AbstractChannel):
         # will unilaterally close the channel and time out the HTLC
         offered_htlc_deadline = lnutil.NBLOCK_DEADLINE_AFTER_EXPIRY_FOR_OFFERED_HTLCS
         for sub, dir, ctn in ((LOCAL, SENT, self.get_latest_ctn(LOCAL)),
-                              (REMOTE, RECEIVED, self.get_oldest_unrevoked_ctn(LOCAL)),
-                              (REMOTE, RECEIVED, self.get_latest_ctn(LOCAL)),):
+                              (REMOTE, RECEIVED, self.get_oldest_unrevoked_ctn(REMOTE)),
+                              (REMOTE, RECEIVED, self.get_latest_ctn(REMOTE)),):
             for htlc_id, htlc in self.hm.htlcs_by_direction(subject=sub, direction=dir, ctn=ctn).items():
                 if htlc.cltv_expiry + offered_htlc_deadline > local_height:
                     continue
@@ -1540,7 +1610,7 @@ class Channel(AbstractChannel):
             return False
         assert conf > 0
         # check funding_tx amount and script
-        funding_tx = self.lnworker.lnwatcher.db.get_transaction(funding_txid)
+        funding_tx = self.lnworker.lnwatcher.adb.get_transaction(funding_txid)
         if not funding_tx:
             self.logger.info(f"no funding_tx {funding_txid}")
             return False
