@@ -12,8 +12,12 @@ from . import util
 from .sql_db import SqlDB, sql
 from .wallet_db import WalletDB
 from .util import bh2u, bfh, log_exceptions, ignore_exceptions, TxMinedInfo, random_shuffled_copy
-from .address_synchronizer import AddressSynchronizer, TX_HEIGHT_LOCAL, TX_HEIGHT_UNCONF_PARENT, TX_HEIGHT_UNCONFIRMED
+from .address_synchronizer import AddressSynchronizer, TX_HEIGHT_LOCAL, TX_HEIGHT_UNCONF_PARENT, TX_HEIGHT_UNCONFIRMED, TX_HEIGHT_FUTURE
 from .transaction import Transaction, TxOutpoint
+from .transaction import match_script_against_template
+from .lnutil import WITNESS_TEMPLATE_RECEIVED_HTLC, WITNESS_TEMPLATE_OFFERED_HTLC
+from .logging import Logger
+
 
 if TYPE_CHECKING:
     from .network import Network
@@ -131,25 +135,25 @@ class SweepStore(SqlDB):
         return [(r[0], r[1]) for r in c.fetchall()]
 
 
+from .util import EventListener, event_listener
+
+class LNWatcher(Logger, EventListener):
 
-class LNWatcher(AddressSynchronizer):
     LOGGING_SHORTCUT = 'W'
 
-    def __init__(self, network: 'Network'):
-        AddressSynchronizer.__init__(self, WalletDB({}, manual_upgrades=False))
+    def __init__(self, adb, network: 'Network'):
+
+        Logger.__init__(self)
+        self.adb = adb
         self.config = network.config
         self.callbacks = {} # address -> lambda: coroutine
         self.network = network
-        util.register_callback(
-            self.on_network_update,
-            ['network_updated', 'blockchain_updated', 'verified', 'wallet_updated', 'fee'])
-
+        self.register_callbacks()
         # status gets populated when we run
         self.channel_status = {}
 
     async def stop(self):
-        await super().stop()
-        util.unregister_callback(self.on_network_update)
+        self.unregister_callbacks()
 
     def get_channel_status(self, outpoint):
         return self.channel_status.get(outpoint, 'unknown')
@@ -168,15 +172,36 @@ class LNWatcher(AddressSynchronizer):
         self.callbacks.pop(address, None)
 
     def add_callback(self, address, callback):
-        self.add_address(address)
+        self.adb.add_address(address)
         self.callbacks[address] = callback
 
+    @event_listener
+    async def on_event_fee(self, *args):
+        await self.trigger_callbacks()
+
+    @event_listener
+    async def on_event_network_updated(self, *args):
+        await self.trigger_callbacks()
+
+    @event_listener
+    async def on_event_blockchain_updated(self, *args):
+        await self.trigger_callbacks()
+
+    @event_listener
+    async def on_event_adb_added_verified_tx(self, adb, tx_hash):
+        if adb != self.adb:
+            return
+        await self.trigger_callbacks()
+
+    @event_listener
+    async def on_event_adb_set_up_to_date(self, adb):
+        if adb != self.adb:
+            return
+        await self.trigger_callbacks()
+
     @log_exceptions
-    async def on_network_update(self, event, *args):
-        if event in ('verified', 'wallet_updated'):
-            if args[0] != self:
-                return
-        if not self.synchronizer:
+    async def trigger_callbacks(self):
+        if not self.adb.synchronizer:
             self.logger.info("synchronizer not set yet")
             return
         for address, callback in list(self.callbacks.items()):
@@ -184,18 +209,18 @@ class LNWatcher(AddressSynchronizer):
 
     async def check_onchain_situation(self, address, funding_outpoint):
         # early return if address has not been added yet
-        if not self.is_mine(address):
+        if not self.adb.is_mine(address):
             return
         spenders = self.inspect_tx_candidate(funding_outpoint, 0)
         # inspect_tx_candidate might have added new addresses, in which case we return ealy
-        if not self.is_up_to_date():
+        if not self.adb.is_up_to_date():
             return
         funding_txid = funding_outpoint.split(':')[0]
-        funding_height = self.get_tx_height(funding_txid)
+        funding_height = self.adb.get_tx_height(funding_txid)
         closing_txid = spenders.get(funding_outpoint)
-        closing_height = self.get_tx_height(closing_txid)
+        closing_height = self.adb.get_tx_height(closing_txid)
         if closing_txid:
-            closing_tx = self.db.get_transaction(closing_txid)
+            closing_tx = self.adb.get_transaction(closing_txid)
             if closing_tx:
                 keep_watching = await self.do_breach_remedy(funding_outpoint, closing_tx, spenders)
             else:
@@ -222,31 +247,60 @@ class LNWatcher(AddressSynchronizer):
         raise NotImplementedError()  # implemented by subclasses
 
     def inspect_tx_candidate(self, outpoint, n):
+        """
+        returns a dict of spenders for a transaction of interest.
+        subscribes to addresses as a side effect.
+        n==0 => outpoint is a channel funding.
+        n==1 => outpoint is a commitment or close output: to_local, to_remote or first-stage htlc
+        n==2 => outpoint is a second-stage htlc
+        """
         prev_txid, index = outpoint.split(':')
-        txid = self.db.get_spent_outpoint(prev_txid, int(index))
-        result = {outpoint:txid}
-        if txid is None:
-            self.channel_status[outpoint] = 'open'
+        spender_txid = self.adb.db.get_spent_outpoint(prev_txid, int(index))
+        result = {outpoint:spender_txid}
+        if n == 0:
+            if spender_txid is None:
+                self.channel_status[outpoint] = 'open'
+            elif not self.is_deeply_mined(spender_txid):
+                self.channel_status[outpoint] = 'closed (%d)' % self.adb.get_tx_height(spender_txid).conf
+            else:
+                self.channel_status[outpoint] = 'closed (deep)'
+        if spender_txid is None:
             return result
-        if n == 0 and not self.is_deeply_mined(txid):
-            self.channel_status[outpoint] = 'closed (%d)' % self.get_tx_height(txid).conf
-        else:
-            self.channel_status[outpoint] = 'closed (deep)'
-        tx = self.db.get_transaction(txid)
-        for i, o in enumerate(tx.outputs()):
+        spender_tx = self.adb.get_transaction(spender_txid)
+        if n == 1:
+            # if tx input is not a first-stage HTLC, we can stop recursion
+            if len(spender_tx.inputs()) != 1:
+                return result
+            o = spender_tx.inputs()[0]
+            witness = o.witness_elements()
+            if not witness:
+                # This can happen if spender_tx is a local unsigned tx in the wallet history, e.g.:
+                # channel is coop-closed, outpoint is for our coop-close output, and spender_tx is an
+                # arbitrary wallet-spend.
+                return result
+            redeem_script = witness[-1]
+            if match_script_against_template(redeem_script, WITNESS_TEMPLATE_OFFERED_HTLC):
+                #self.logger.info(f"input script matches offered htlc {redeem_script.hex()}")
+                pass
+            elif match_script_against_template(redeem_script, WITNESS_TEMPLATE_RECEIVED_HTLC):
+                #self.logger.info(f"input script matches received htlc {redeem_script.hex()}")
+                pass
+            else:
+                return result
+        for i, o in enumerate(spender_tx.outputs()):
             if o.address is None:
                 continue
-            if not self.is_mine(o.address):
-                self.add_address(o.address)
+            if not self.adb.is_mine(o.address):
+                self.adb.add_address(o.address)
             elif n < 2:
-                r = self.inspect_tx_candidate(txid+':%d'%i, n+1)
+                r = self.inspect_tx_candidate(spender_txid+':%d'%i, n+1)
                 result.update(r)
         return result
 
     def get_tx_mined_depth(self, txid: str):
         if not txid:
             return TxMinedDepth.FREE
-        tx_mined_depth = self.get_tx_height(txid)
+        tx_mined_depth = self.adb.get_tx_height(txid)
         height, conf = tx_mined_depth.height, tx_mined_depth.conf
         if conf > 100:
             return TxMinedDepth.DEEP
@@ -254,7 +308,7 @@ class LNWatcher(AddressSynchronizer):
             return TxMinedDepth.SHALLOW
         elif height in (TX_HEIGHT_UNCONFIRMED, TX_HEIGHT_UNCONF_PARENT):
             return TxMinedDepth.MEMPOOL
-        elif height == TX_HEIGHT_LOCAL:
+        elif height in (TX_HEIGHT_LOCAL, TX_HEIGHT_FUTURE):
             return TxMinedDepth.FREE
         elif height > 0 and conf == 0:
             # unverified but claimed to be mined
@@ -271,13 +325,19 @@ class WatchTower(LNWatcher):
     LOGGING_SHORTCUT = 'W'
 
     def __init__(self, network):
-        LNWatcher.__init__(self, network)
+        adb = AddressSynchronizer(WalletDB({}, manual_upgrades=False), network.config)
+        adb.start_network(network)
+        LNWatcher.__init__(self, adb, network)
         self.network = network
         self.sweepstore = SweepStore(os.path.join(self.network.config.path, "watchtower_db"), network)
         # this maps funding_outpoints to ListenerItems, which have an event for when the watcher is done,
         # and a queue for seeing which txs are being published
         self.tx_progress = {} # type: Dict[str, ListenerItem]
 
+    async def stop(self):
+        await super().stop()
+        await self.adb.stop()
+
     def diagnostic_name(self):
         return "local_tower"
 
@@ -300,7 +360,7 @@ class WatchTower(LNWatcher):
         return keep_watching
 
     async def broadcast_or_log(self, funding_outpoint: str, tx: Transaction):
-        height = self.get_tx_height(tx.txid()).height
+        height = self.adb.get_tx_height(tx.txid()).height
         if height != TX_HEIGHT_LOCAL:
             return
         try:
@@ -352,7 +412,7 @@ class LNWalletWatcher(LNWatcher):
     def __init__(self, lnworker: 'LNWallet', network: 'Network'):
         self.network = network
         self.lnworker = lnworker
-        LNWatcher.__init__(self, network)
+        LNWatcher.__init__(self, lnworker.wallet.adb, network)
 
     def diagnostic_name(self):
         return f"{self.lnworker.wallet.diagnostic_name()}-LNW"
@@ -365,13 +425,15 @@ class LNWalletWatcher(LNWatcher):
         chan = self.lnworker.channel_by_txo(funding_outpoint)
         if not chan:
             return
-        chan.update_onchain_state(funding_txid=funding_txid,
-                                  funding_height=funding_height,
-                                  closing_txid=closing_txid,
-                                  closing_height=closing_height,
-                                  keep_watching=keep_watching)
-        await self.lnworker.on_channel_update(chan)
+        chan.update_onchain_state(
+            funding_txid=funding_txid,
+            funding_height=funding_height,
+            closing_txid=closing_txid,
+            closing_height=closing_height,
+            keep_watching=keep_watching)
+        await self.lnworker.handle_onchain_state(chan)
 
+    @log_exceptions
     async def do_breach_remedy(self, funding_outpoint, closing_tx, spenders):
         chan = self.lnworker.channel_by_txo(funding_outpoint)
         if not chan:
@@ -384,73 +446,100 @@ class LNWalletWatcher(LNWatcher):
         for prevout, sweep_info in sweep_info_dict.items():
             name = sweep_info.name + ' ' + chan.get_id_for_log()
             spender_txid = spenders.get(prevout)
-            if spender_txid is not None:
-                spender_tx = self.db.get_transaction(spender_txid)
-                if not spender_tx:
-                    keep_watching = True
-                    continue
+            spender_tx = self.adb.get_transaction(spender_txid) if spender_txid else None
+            if spender_tx:
+                # the spender might be the remote, revoked or not
                 e_htlc_tx = chan.maybe_sweep_revoked_htlc(closing_tx, spender_tx)
                 if e_htlc_tx:
                     spender2 = spenders.get(spender_txid+':0')
                     if spender2:
                         keep_watching |= not self.is_deeply_mined(spender2)
                     else:
-                        await self.try_redeem(spender_txid+':0', e_htlc_tx, chan_id_for_log, name)
                         keep_watching = True
+                    await self.maybe_redeem(spenders, spender_txid+':0', e_htlc_tx, name)
                 else:
-                    keep_watching |= not self.is_deeply_mined(spender_txid)
+                    keep_watching |= not self.is_deeply_mined(spender_tx.txid())
                     txin_idx = spender_tx.get_input_idx_that_spent_prevout(TxOutpoint.from_str(prevout))
                     assert txin_idx is not None
                     spender_txin = spender_tx.inputs()[txin_idx]
                     chan.extract_preimage_from_htlc_txin(spender_txin)
             else:
-                await self.try_redeem(prevout, sweep_info, chan_id_for_log, name)
                 keep_watching = True
+            # broadcast or maybe update our own tx
+            await self.maybe_redeem(spenders, prevout, sweep_info, name)
+
         return keep_watching
 
-    @log_exceptions
-    async def try_redeem(self, prevout: str, sweep_info: 'SweepInfo', chan_id_for_log: str, name: str) -> None:
+    def get_redeem_tx(self, spenders, prevout: str, sweep_info: 'SweepInfo', name: str):
+        # check if redeem tx needs to be updated
+        # if it is in the mempool, we need to check fee rise
+        txid = spenders.get(prevout)
+        old_tx = self.adb.get_transaction(txid)
+        assert old_tx is not None or txid is None
+        tx_depth = self.get_tx_mined_depth(txid) if txid else None
+        if txid and tx_depth not in [TxMinedDepth.FREE, TxMinedDepth.MEMPOOL]:
+            assert old_tx is not None
+            return old_tx, None
+        new_tx = sweep_info.gen_tx()
+        if new_tx is None:
+            self.logger.info(f'{name} could not claim output: {prevout}, dust')
+            assert old_tx is not None
+            return old_tx, None
+        if txid is None:
+            return None, new_tx
+        elif tx_depth == TxMinedDepth.MEMPOOL:
+            delta = new_tx.get_fee() - self.adb.get_tx_fee(txid)
+            if delta > 1:
+                self.logger.info(f'increasing fee of mempool tx {name}: {prevout}')
+                return old_tx, new_tx
+            else:
+                assert old_tx is not None
+                return old_tx, None
+        elif tx_depth == TxMinedDepth.FREE:
+            # return new tx, even if it is equal to old_tx,
+            # because we need to test if it can be broadcast
+            return old_tx, new_tx
+        else:
+            assert old_tx is not None
+            return old_tx, None
+
+    async def maybe_redeem(self, spenders, prevout, sweep_info: 'SweepInfo', name: str) -> None:
+        old_tx, new_tx = self.get_redeem_tx(spenders, prevout, sweep_info, name)
+        if new_tx is None:
+            return
         prev_txid, prev_index = prevout.split(':')
-        broadcast = True
+        can_broadcast = True
         local_height = self.network.get_local_height()
         if sweep_info.cltv_expiry:
-            wanted_height = sweep_info.cltv_expiry - local_height
+            wanted_height = sweep_info.cltv_expiry
             if wanted_height - local_height > 0:
-                broadcast = False
-                reason = 'waiting for {}: CLTV ({} > {}), prevout {}'.format(name, local_height, sweep_info.cltv_expiry, prevout)
+                can_broadcast = False
+                reason = 'waiting for {}: CLTV ({} > {})'.format(name, local_height, sweep_info.cltv_expiry)
         if sweep_info.csv_delay:
-            prev_height = self.get_tx_height(prev_txid)
+            prev_height = self.adb.get_tx_height(prev_txid)
             wanted_height = sweep_info.csv_delay + prev_height.height - 1
-            if wanted_height - local_height > 0:
-                broadcast = False
-                reason = 'waiting for {}: CSV ({} >= {}), prevout: {}'.format(name, prev_height.conf, sweep_info.csv_delay, prevout)
-        tx = sweep_info.gen_tx()
-        if tx is None:
-            self.logger.info(f'{name} could not claim output: {prevout}, dust')
-            return
-        txid = tx.txid()
-        self.lnworker.wallet.set_label(txid, name)
-        if broadcast:
-            await self.network.try_broadcasting(tx, name)
+            if prev_height.height <= 0 or wanted_height - local_height > 0:
+                can_broadcast = False
+                reason = 'waiting for {}: CSV ({} >= {})'.format(name, prev_height.conf, sweep_info.csv_delay)
+        if can_broadcast:
+            self.logger.info(f'we can broadcast: {name}')
+            tx_was_added = await self.network.try_broadcasting(new_tx, name)
         else:
-            if txid in self.lnworker.wallet.future_tx:
-                return
-            self.logger.debug(f'(chan {chan_id_for_log}) trying to redeem {name}: {prevout}')
-            self.logger.info(reason)
-            # it's OK to add local transaction, the fee will be recomputed
-            try:
-                tx_was_added = self.lnworker.wallet.add_future_tx(tx, wanted_height)
-            except Exception as e:
-                self.logger.info(f'could not add future tx: {name}. prevout: {prevout} {str(e)}')
+            self.logger.info(f'cannot broadcast: {name} {reason}')
+            # we may have a tx with a different fee, in which case it will be replaced
+            if old_tx and old_tx.txid() != new_tx.txid():
+                try:
+                    tx_was_added = self.adb.add_transaction(new_tx, notify_GUI=(old_tx is None))
+                except Exception as e:
+                    self.logger.info(f'could not add future tx: {name}. prevout: {prevout} {str(e)}')
+                    tx_was_added = False
+                if tx_was_added:
+                    self.adb.set_future_tx(new_tx.txid(), wanted_height)
+                    self.logger.info(f'added redeem tx: {name}. prevout: {prevout}')
+            else:
                 tx_was_added = False
-            if tx_was_added:
-                self.logger.info(f'added future tx: {name}. prevout: {prevout}')
-                util.trigger_callback('wallet_updated', self.lnworker.wallet)
-
-    def add_verified_tx(self, tx_hash: str, info: TxMinedInfo):
-        # this method is overloaded so that we have the GUI refreshed
-        # TODO: LNWatcher should not be an AddressSynchronizer,
-        # we should use the existing wallet instead, and results would be persisted
-        super().add_verified_tx(tx_hash, info)
-        tx_mined_status = self.get_tx_height(tx_hash)
-        util.trigger_callback('verified', self.lnworker.wallet, tx_hash, tx_mined_status)
+        if tx_was_added:
+            self.lnworker.wallet.set_label(new_tx.txid(), name)
+            if old_tx and old_tx.txid() != new_tx.txid():
+                self.lnworker.wallet.set_label(old_tx.txid(), None)
+            util.trigger_callback('wallet_updated', self.lnworker.wallet)
