@@ -9,7 +9,7 @@ from PyQt5.QtWidgets import (QVBoxLayout, QLabel, QGridLayout, QPushButton,
                              QMessageBox, QFileDialog, QSlider, QTabWidget)
 
 from electrum.gui.qt.util import (WindowModalDialog, WWLabel, Buttons, CancelButton,
-                                  OkButton, CloseButton)
+                                  OkButton, CloseButton, getOpenFileName)
 from electrum.i18n import _
 from electrum.plugin import hook
 from electrum.util import bh2u
@@ -38,24 +38,24 @@ PASSPHRASE_NOT_PIN = _(
 
 class QtHandler(QtHandlerBase):
 
-    pin_signal = pyqtSignal(object)
+    pin_signal = pyqtSignal(object, object)
 
     def __init__(self, win, pin_matrix_widget_class, device):
         super(QtHandler, self).__init__(win, device)
         self.pin_signal.connect(self.pin_dialog)
         self.pin_matrix_widget_class = pin_matrix_widget_class
 
-    def get_pin(self, msg):
+    def get_pin(self, msg, *, show_strength=True):
         self.done.clear()
-        self.pin_signal.emit(msg)
+        self.pin_signal.emit(msg, show_strength)
         self.done.wait()
         return self.response
 
-    def pin_dialog(self, msg):
+    def pin_dialog(self, msg, show_strength):
         # Needed e.g. when resetting a device
         self.clear_dialog()
         dialog = WindowModalDialog(self.top_level_window(), _("Enter PIN"))
-        matrix = self.pin_matrix_widget_class()
+        matrix = self.pin_matrix_widget_class(show_strength)
         vbox = QVBoxLayout()
         vbox.addWidget(QLabel(msg))
         vbox.addWidget(matrix)
@@ -71,9 +71,6 @@ class QtPlugin(QtPluginBase):
     #   icon_file
     #   pin_matrix_widget_class
 
-    def create_handler(self, window):
-        return QtHandler(window, self.pin_matrix_widget_class(), self.device)
-
     @only_hook_if_libraries_available
     @hook
     def receive_menu(self, menu, addrs, wallet):
@@ -87,9 +84,13 @@ class QtPlugin(QtPluginBase):
                 menu.addAction(_("Show on {}").format(device_name), show_address)
 
     def show_settings_dialog(self, window, keystore):
-        device_id = self.choose_device(window, keystore)
-        if device_id:
-            SettingsDialog(window, self, keystore, device_id).exec_()
+        def connect():
+            device_id = self.choose_device(window, keystore)
+            return device_id
+        def show_dialog(device_id):
+            if device_id:
+                SettingsDialog(window, self, keystore, device_id).exec_()
+        keystore.thread.add(connect, on_success=show_dialog)
 
     def request_safe_t_init_settings(self, wizard, method, device):
         vbox = QVBoxLayout()
@@ -176,12 +177,12 @@ class Plugin(SafeTPlugin, QtPlugin):
     icon_unpaired = "safe-t_unpaired.png"
     icon_paired = "safe-t.png"
 
+    def create_handler(self, window):
+        return QtHandler(window, self.pin_matrix_widget_class(), self.device)
+
     @classmethod
     def pin_matrix_widget_class(self):
-        # We use a local updated copy of pinmatrix.py until safetlib
-        # releases a new version that includes https://github.com/archos-safe-t/python-safet/commit/b1eab3dba4c04fdfc1fcf17b66662c28c5f2380e
-        # from safetlib.qt.pinmatrix import PinMatrixWidget
-        from .pinmatrix import PinMatrixWidget
+        from safetlib.qt.pinmatrix import PinMatrixWidget
         return PinMatrixWidget
 
 
@@ -275,8 +276,11 @@ class SettingsDialog(WindowModalDialog):
             invoke_client('toggle_passphrase', unpair_after=currently_enabled)
 
         def change_homescreen():
-            dialog = QFileDialog(self, _("Choose Homescreen"))
-            filename, __ = dialog.getOpenFileName()
+            filename = getOpenFileName(
+                parent=self,
+                title=_("Choose Homescreen"),
+                config=config,
+            )
             if not filename:
                 return  # user cancelled
 
